@@ -274,3 +274,8 @@ pub(crate) const fn is_fix_write_whole_file_enabled(settings: &LinterSettings) -
 pub(crate) const fn is_typing_extensions_str_alias_enabled(settings: &LinterSettings) -> bool {
     settings.preview.is_enabled()
 }
+
+// https://github.com/astral-sh/ruff/pull/19045
+pub(crate) const fn is_extended_i18n_function_matching_enabled(settings: &LinterSettings) -> bool {
+    settings.preview.is_enabled()
+}
