@@ -52,6 +52,7 @@ mod tests {
     #[test_case(Rule::ManualFromImport, Path::new("import_aliasing.py"))]
     #[test_case(Rule::IfStmtMinMax, Path::new("if_stmt_min_max.py"))]
     #[test_case(Rule::SingleStringSlots, Path::new("single_string_slots.py"))]
+    #[test_case(Rule::StopIterationReturn, Path::new("stop_iteration_return.py"))]
     #[test_case(Rule::SysExitAlias, Path::new("sys_exit_alias_0.py"))]
     #[test_case(Rule::SysExitAlias, Path::new("sys_exit_alias_1.py"))]
     #[test_case(Rule::SysExitAlias, Path::new("sys_exit_alias_2.py"))]
@@ -252,30 +253,6 @@ mod tests {
         Ok(())
     }
 
-    #[test_case(Rule::BidirectionalUnicode, Path::new("bidirectional_unicode.py"))]
-    fn preview_rules(rule_code: Rule, path: &Path) -> Result<()> {
-        let snapshot = format!(
-            "preview__{}_{}",
-            rule_code.noqa_code(),
-            path.to_string_lossy()
-        );
-        let diagnostics = test_path(
-            Path::new("pylint").join(path).as_path(),
-            &LinterSettings {
-                pylint: pylint::settings::Settings {
-                    allow_dunder_method_names: FxHashSet::from_iter([
-                        "__special_custom_magic__".to_string()
-                    ]),
-                    ..pylint::settings::Settings::default()
-                },
-                preview: PreviewMode::Enabled,
-                ..LinterSettings::for_rule(rule_code)
-            },
-        )?;
-        assert_diagnostics!(snapshot, diagnostics);
-        Ok(())
-    }
-
     #[test]
     fn continue_in_finally() -> Result<()> {
         let diagnostics = test_path(
@@ -444,19 +421,6 @@ mod tests {
         Ok(())
     }
 
-    #[test]
-    fn preview_useless_import_alias() -> Result<()> {
-        let diagnostics = test_path(
-            Path::new("pylint/import_aliasing_2/__init__.py"),
-            &LinterSettings {
-                preview: PreviewMode::Enabled,
-                ..LinterSettings::for_rule(Rule::UselessImportAlias)
-            },
-        )?;
-        assert_diagnostics!(diagnostics);
-        Ok(())
-    }
-
     #[test]
     fn import_outside_top_level_with_banned() -> Result<()> {
         let diagnostics = test_path(
