@@ -1,3 +1,4 @@
+pub(crate) use access_annotations_from_class_dict::*;
 pub(crate) use ambiguous_unicode_character::*;
 pub(crate) use assert_with_print_message::*;
 pub(crate) use assignment_in_assert::*;
@@ -22,6 +23,7 @@ pub(crate) use invalid_index_type::*;
 pub(crate) use invalid_pyproject_toml::*;
 pub(crate) use invalid_rule_code::*;
 pub(crate) use legacy_form_pytest_raises::*;
+pub(crate) use logging_eager_conversion::*;
 pub(crate) use map_int_version_parsing::*;
 pub(crate) use missing_fstring_syntax::*;
 pub(crate) use mutable_class_default::*;
@@ -29,6 +31,7 @@ pub(crate) use mutable_dataclass_default::*;
 pub(crate) use mutable_fromkeys_value::*;
 pub(crate) use needless_else::*;
 pub(crate) use never_union::*;
+pub(crate) use non_octal_permissions::*;
 pub(crate) use none_not_at_end_of_union::*;
 pub(crate) use parenthesize_chained_operators::*;
 pub(crate) use post_init_default::*;
@@ -58,6 +61,7 @@ pub(crate) use used_dummy_variable::*;
 pub(crate) use useless_if_else::*;
 pub(crate) use zip_instead_of_pairwise::*;
 
+mod access_annotations_from_class_dict;
 mod ambiguous_unicode_character;
 mod assert_with_print_message;
 mod assignment_in_assert;
@@ -71,7 +75,6 @@ mod default_factory_kwarg;
 mod explicit_f_string_type_conversion;
 mod falsy_dict_get_fallback;
 mod function_call_in_dataclass_default;
-mod helpers;
 mod if_key_in_dict_del;
 mod implicit_classvar_in_dataclass;
 mod implicit_optional;
@@ -84,6 +87,7 @@ mod invalid_index_type;
 mod invalid_pyproject_toml;
 mod invalid_rule_code;
 mod legacy_form_pytest_raises;
+mod logging_eager_conversion;
 mod map_int_version_parsing;
 mod missing_fstring_syntax;
 mod mutable_class_default;
@@ -91,6 +95,7 @@ mod mutable_dataclass_default;
 mod mutable_fromkeys_value;
 mod needless_else;
 mod never_union;
+mod non_octal_permissions;
 mod none_not_at_end_of_union;
 mod parenthesize_chained_operators;
 mod post_init_default;
