@@ -1,45 +1,53 @@
 //! Detect code style from Python source code.
 
+use std::borrow::Cow;
+use std::cell::OnceCell;
 use std::ops::Deref;
 
-use once_cell::unsync::OnceCell;
-
 use ruff_python_ast::str::Quote;
 use ruff_python_parser::{Token, TokenKind, Tokens};
-use ruff_source_file::{find_newline, LineEnding, Locator};
+use ruff_source_file::{LineEnding, LineRanges, find_newline};
 use ruff_text_size::Ranged;
 
 #[derive(Debug, Clone)]
 pub struct Stylist<'a> {
-    locator: &'a Locator<'a>,
+    source: Cow<'a, str>,
     indentation: Indentation,
     quote: Quote,
     line_ending: OnceCell<LineEnding>,
 }
 
 impl<'a> Stylist<'a> {
-    pub fn indentation(&'a self) -> &'a Indentation {
+    pub fn indentation(&self) -> &Indentation {
         &self.indentation
     }
 
-    pub fn quote(&'a self) -> Quote {
+    pub fn quote(&self) -> Quote {
         self.quote
     }
 
-    pub fn line_ending(&'a self) -> LineEnding {
+    pub fn line_ending(&self) -> LineEnding {
         *self.line_ending.get_or_init(|| {
-            let contents = self.locator.contents();
-            find_newline(contents)
+            find_newline(&self.source)
                 .map(|(_, ending)| ending)
                 .unwrap_or_default()
         })
     }
 
-    pub fn from_tokens(tokens: &Tokens, locator: &'a Locator<'a>) -> Self {
-        let indentation = detect_indentation(tokens, locator);
+    pub fn into_owned(self) -> Stylist<'static> {
+        Stylist {
+            source: Cow::Owned(self.source.into_owned()),
+            indentation: self.indentation,
+            quote: self.quote,
+            line_ending: self.line_ending,
+        }
+    }
+
+    pub fn from_tokens(tokens: &Tokens, source: &'a str) -> Self {
+        let indentation = detect_indentation(tokens, source);
 
         Self {
-            locator,
+            source: Cow::Borrowed(source),
             indentation,
             quote: detect_quote(tokens),
             line_ending: OnceCell::default(),
@@ -51,7 +59,7 @@ fn detect_quote(tokens: &[Token]) -> Quote {
     for token in tokens {
         match token.kind() {
             TokenKind::String if !token.is_triple_quoted_string() => {
-                return token.string_quote_style()
+                return token.string_quote_style();
             }
             TokenKind::FStringStart => return token.string_quote_style(),
             _ => continue,
@@ -60,7 +68,7 @@ fn detect_quote(tokens: &[Token]) -> Quote {
     Quote::default()
 }
 
-fn detect_indentation(tokens: &[Token], locator: &Locator) -> Indentation {
+fn detect_indentation(tokens: &[Token], source: &str) -> Indentation {
     let indent_range = tokens.iter().find_map(|token| {
         if matches!(token.kind(), TokenKind::Indent) {
             Some(token.range())
@@ -70,7 +78,7 @@ fn detect_indentation(tokens: &[Token], locator: &Locator) -> Indentation {
     });
 
     if let Some(indent_range) = indent_range {
-        let mut whitespace = locator.slice(indent_range);
+        let mut whitespace = &source[indent_range];
         // https://docs.python.org/3/reference/lexical_analysis.html#indentation
         // > A formfeed character may be present at the start of the line; it will be ignored for
         // > the indentation calculations above. Formfeed characters occurring elsewhere in the
@@ -99,7 +107,7 @@ fn detect_indentation(tokens: &[Token], locator: &Locator) -> Indentation {
         // ```
         for token in tokens {
             if token.kind() == TokenKind::NonLogicalNewline {
-                let line = locator.line(token.end());
+                let line = source.line_str(token.end());
                 let indent_index = line.find(|c: char| !c.is_whitespace());
                 if let Some(indent_index) = indent_index {
                     if indent_index > 0 {
@@ -150,46 +158,40 @@ impl Deref for Indentation {
 
 #[cfg(test)]
 mod tests {
-    use ruff_python_parser::{parse_module, parse_unchecked, Mode};
-
-    use ruff_source_file::{find_newline, LineEnding};
+    use ruff_python_parser::{Mode, ParseOptions, parse_module, parse_unchecked};
+    use ruff_source_file::{LineEnding, find_newline};
 
     use super::{Indentation, Quote, Stylist};
-    use ruff_source_file::Locator;
 
     #[test]
     fn indentation() {
         let contents = r"x = 1";
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.indentation(), &Indentation::default());
 
         let contents = r"
 if True:
   pass
 ";
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.indentation(), &Indentation("  ".to_string()));
 
         let contents = r"
 if True:
     pass
 ";
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.indentation(), &Indentation("    ".to_string()));
 
         let contents = r"
 if True:
 	pass
 ";
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.indentation(), &Indentation("\t".to_string()));
 
         let contents = r"
@@ -199,9 +201,8 @@ x = (
   3,
 )
 ";
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.indentation(), &Indentation("  ".to_string()));
 
         // formfeed indent, see `detect_indentation` comment.
@@ -210,9 +211,8 @@ class FormFeedIndent:
    def __init__(self, a=[]):
         print(a)
 ";
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.indentation(), &Indentation(" ".to_string()));
     }
 
@@ -225,10 +225,9 @@ x = (
  3,
 )
 ";
-        let locator = Locator::new(contents);
-        let parsed = parse_unchecked(contents, Mode::Module);
+        let parsed = parse_unchecked(contents, ParseOptions::from(Mode::Module));
         assert_eq!(
-            Stylist::from_tokens(parsed.tokens(), &locator).indentation(),
+            Stylist::from_tokens(parsed.tokens(), contents).indentation(),
             &Indentation(" ".to_string())
         );
     }
@@ -236,39 +235,33 @@ x = (
     #[test]
     fn quote() {
         let contents = r"x = 1";
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.quote(), Quote::default());
 
         let contents = r"x = '1'";
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.quote(), Quote::Single);
 
         let contents = r"x = f'1'";
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.quote(), Quote::Single);
 
         let contents = r#"x = "1""#;
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.quote(), Quote::Double);
 
         let contents = r#"x = f"1""#;
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.quote(), Quote::Double);
 
         let contents = r#"s = "It's done.""#;
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.quote(), Quote::Double);
 
         // No style if only double quoted docstring (will take default Double)
@@ -277,9 +270,8 @@ def f():
     """Docstring."""
     pass
 "#;
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.quote(), Quote::default());
 
         // Detect from string literal appearing after docstring
@@ -288,9 +280,8 @@ def f():
 
 a = 'v'
 "#;
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.quote(), Quote::Single);
 
         let contents = r#"
@@ -298,9 +289,8 @@ a = 'v'
 
 a = "v"
 "#;
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.quote(), Quote::Double);
 
         // Detect from f-string appearing after docstring
@@ -309,9 +299,8 @@ a = "v"
 
 a = f'v'
 "#;
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.quote(), Quote::Single);
 
         let contents = r#"
@@ -319,17 +308,15 @@ a = f'v'
 
 a = f"v"
 "#;
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.quote(), Quote::Double);
 
         let contents = r"
 f'''Module docstring.'''
 ";
-        let locator = Locator::new(contents);
         let parsed = parse_module(contents).unwrap();
-        let stylist = Stylist::from_tokens(parsed.tokens(), &locator);
+        let stylist = Stylist::from_tokens(parsed.tokens(), contents);
         assert_eq!(stylist.quote(), Quote::Single);
     }
 
