@@ -1,6 +1,6 @@
 use std::num::NonZeroUsize;
 
-use crate::session::Session;
+use crate::session::{Client, Session};
 
 mod task;
 mod thread;
@@ -12,13 +12,11 @@ use self::{
     thread::ThreadPriority,
 };
 
-use super::{client::Client, ClientSender};
-
 /// The event loop thread is actually a secondary thread that we spawn from the
 /// _actual_ main thread. This secondary thread has a larger stack size
 /// than some OS defaults (Windows, for example) and is also designated as
 /// high-priority.
-pub(crate) fn event_loop_thread(
+pub(crate) fn spawn_main_loop(
     func: impl FnOnce() -> crate::Result<()> + Send + 'static,
 ) -> crate::Result<thread::JoinHandle<crate::Result<()>>> {
     // Override OS defaults to avoid stack overflows on platforms with low stack size defaults.
@@ -32,69 +30,33 @@ pub(crate) fn event_loop_thread(
     )
 }
 
-pub(crate) struct Scheduler<'s> {
-    session: &'s mut Session,
-    client: Client<'s>,
+pub(crate) struct Scheduler {
     fmt_pool: thread::Pool,
     background_pool: thread::Pool,
 }
 
-impl<'s> Scheduler<'s> {
-    pub(super) fn new(
-        session: &'s mut Session,
-        worker_threads: NonZeroUsize,
-        sender: ClientSender,
-    ) -> Self {
+impl Scheduler {
+    pub(super) fn new(worker_threads: NonZeroUsize) -> Self {
         const FMT_THREADS: usize = 1;
         Self {
-            session,
             fmt_pool: thread::Pool::new(NonZeroUsize::try_from(FMT_THREADS).unwrap()),
             background_pool: thread::Pool::new(worker_threads),
-            client: Client::new(sender),
         }
     }
 
-    /// Immediately sends a request of kind `R` to the client, with associated parameters.
-    /// The task provided by `response_handler` will be dispatched as soon as the response
-    /// comes back from the client.
-    pub(super) fn request<R>(
-        &mut self,
-        params: R::Params,
-        response_handler: impl Fn(R::Result) -> Task<'s> + 'static,
-    ) -> crate::Result<()>
-    where
-        R: lsp_types::request::Request,
-    {
-        self.client.requester.request::<R>(params, response_handler)
-    }
-
-    /// Creates a task to handle a response from the client.
-    pub(super) fn response(&mut self, response: lsp_server::Response) -> Task<'s> {
-        self.client.requester.pop_response_task(response)
-    }
-
     /// Dispatches a `task` by either running it as a blocking function or
     /// executing it on a background thread pool.
-    pub(super) fn dispatch(&mut self, task: task::Task<'s>) {
+    pub(super) fn dispatch(&mut self, task: Task, session: &mut Session, client: Client) {
         match task {
             Task::Sync(SyncTask { func }) => {
-                let notifier = self.client.notifier();
-                let responder = self.client.responder();
-                func(
-                    self.session,
-                    notifier,
-                    &mut self.client.requester,
-                    responder,
-                );
+                func(session, &client);
             }
             Task::Background(BackgroundTaskBuilder {
                 schedule,
                 builder: func,
             }) => {
-                let static_func = func(self.session);
-                let notifier = self.client.notifier();
-                let responder = self.client.responder();
-                let task = move || static_func(notifier, responder);
+                let static_func = func(session);
+                let task = move || static_func(&client);
                 match schedule {
                     BackgroundSchedule::Worker => {
                         self.background_pool.spawn(ThreadPriority::Worker, task);
